@@ -25,6 +25,11 @@ struct  _EogIter {
 	GList         *node;
 };
 
+typedef struct {
+	GnomeVFSMonitorHandle *handle;
+	const gchar *text_uri;
+} MonitorHandleContext;
+
 struct _EogImageListPrivate {
 	/* The age is incremented everytime the list is altered. 
 	   Iterators are only valid if they have the same age */
@@ -36,9 +41,15 @@ struct _EogImageListPrivate {
 	/* Number of images. Same as g_list_length (store) */
 	int             n_images;
 
+	/* The initial image pos */
+	int            initial_pos;
+
 	/* the largest common part of all image uris */
 	GnomeVFSURI     *base_uri;
 	gboolean         no_common_base;
+	
+	GList *monitors;
+
 };
 
 typedef struct {
@@ -104,6 +115,16 @@ eog_image_list_finalize (GObject *object)
 	}
 }
 
+static void
+foreach_monitors_free (gpointer data, gpointer user_data)
+{
+	MonitorHandleContext *hctx = data;
+	
+	gnome_vfs_monitor_cancel (hctx->handle);
+	
+	g_free (data);
+}
+
 static void
 eog_image_list_dispose (GObject *object)
 {
@@ -130,10 +151,14 @@ eog_image_list_dispose (GObject *object)
 		gnome_vfs_uri_unref (priv->base_uri);
 		priv->base_uri = NULL;
 	}
+	
+	g_list_foreach (priv->monitors, 
+		foreach_monitors_free, NULL);
+	g_list_free (priv->monitors);	
 }
 
 static void
-eog_image_list_instance_init (EogImageList *obj)
+eog_image_list_init (EogImageList *obj)
 {
 	EogImageListPrivate *priv;
 
@@ -146,6 +171,9 @@ eog_image_list_instance_init (EogImageList *obj)
 	priv->n_images = 0;
 	priv->base_uri = NULL;
 	priv->no_common_base = FALSE;
+	priv->initial_pos = -1;
+	
+	priv->monitors = NULL;
 }
 
 static void 
@@ -192,14 +220,28 @@ eog_image_list_class_init (EogImageListClass *klass)
 }
 
 
-GNOME_CLASS_BOILERPLATE (EogImageList,
-			 eog_image_list,
-			 GObject,
-			 G_TYPE_OBJECT);
+G_DEFINE_TYPE (EogImageList, eog_image_list, G_TYPE_OBJECT)
 
 
 /* ================== Private EogImageList functions  ===================*/
 
+static int
+compare_uri_cb (gconstpointer element_data, gconstpointer data)
+{
+	EogImage *image;
+	GnomeVFSURI *uri;
+
+	image = EOG_IMAGE (element_data);
+	uri = (GnomeVFSURI*) data;
+
+	if (gnome_vfs_uri_equal (eog_image_get_uri (image), uri)) {
+		return 0;
+	} 
+	else {
+		return 1;
+	}
+}
+
 static int
 compare_filename_cb (gconstpointer a, gconstpointer b)
 {
@@ -365,6 +407,96 @@ is_supported_mime_type (const char *mime_type)
 
 /* ================== Directory Loading stuff ===================*/
 
+
+static void 
+cleanup_dead_files (EogImageList *list)
+{
+	gint position = 0;
+	GList *node = NULL, *remove = NULL;
+	gboolean found = FALSE;
+	
+	if (list && list->priv)
+		node = list->priv->store;
+
+	while (node != NULL && !found) {
+		GnomeVFSURI *uri = eog_image_get_uri(node->data);
+		if (!gnome_vfs_uri_exists (uri)) {
+			remove = g_list_append (remove, (gpointer)position);
+		}	
+		node = node->next;		
+		position++;
+	}
+
+	while (remove) {
+		gint pos = (gint)remove->data;
+		EogIter *iter = eog_image_list_get_iter_by_pos (list, pos);
+
+		if (iter) {
+			remove_image_private (list, iter);
+			g_signal_emit (list, eog_image_list_signals[IMAGE_REMOVED], 0, pos);
+		}
+		
+		remove = g_list_next (remove);
+	}
+	
+	g_list_free (remove);
+}
+
+static void
+vfs_monitor_dir_cb (GnomeVFSMonitorHandle *handle,
+          const gchar *monitor_uri,
+          const gchar *info_uri,
+          GnomeVFSMonitorEventType event_type,
+          gpointer user_data)
+{
+	EogImageList *list = user_data;
+	GnomeVFSURI *uri = NULL;
+	EogImage *image = NULL;
+	GList *node=NULL;
+	gboolean found = FALSE;
+	
+	cleanup_dead_files (list);
+	
+	switch (event_type) {
+		case GNOME_VFS_MONITOR_EVENT_CHANGED:
+		case GNOME_VFS_MONITOR_EVENT_METADATA_CHANGED:
+		case GNOME_VFS_MONITOR_EVENT_DELETED:
+		case GNOME_VFS_MONITOR_EVENT_STARTEXECUTING:
+		case GNOME_VFS_MONITOR_EVENT_STOPEXECUTING:
+			break;
+			
+		case GNOME_VFS_MONITOR_EVENT_CREATED:
+		
+			if (list && list->priv) {
+				node = list->priv->store;
+				
+				while (node != NULL && !found) {
+			
+					uri = eog_image_get_uri(node->data);
+					gchar *str = gnome_vfs_uri_to_string 
+						(uri, GNOME_VFS_URI_HIDE_NONE);
+					found = (strcmp (str, info_uri)==0)?TRUE:FALSE;
+					g_free (str);
+					node = node->next;
+				}
+			}
+			
+			if (!found) {
+				uri = gnome_vfs_uri_new (info_uri);
+				image = eog_image_new_uri (uri);
+				
+				if (image) {
+					eog_image_list_add_image (list, image);
+					/* Q: g_object_unref (image) ? */
+				}
+			}
+			gnome_vfs_uri_unref (uri);
+			break;
+	}
+	
+	return;	
+}
+
 /* Called for each file in a directory. Checks if the file is some
  * sort of image. If so, it creates an image object and adds it to the
  * list.
@@ -379,13 +511,11 @@ directory_visit_cb (const gchar *rel_uri,
 	EogImage *image;
 	GnomeVFSURI *uri;
 	EogImageList *list;
-	EogImageListPrivate *priv;
 	gboolean load_uri = FALSE;
 	DirLoadingContext *ctx;
 	
 	ctx = (DirLoadingContext*) data;
 	list = ctx->list;
-	priv = list->priv;
 
 	if ((info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE) > 0) {
 		if (is_supported_mime_type (info->mime_type)) {
@@ -394,12 +524,10 @@ directory_visit_cb (const gchar *rel_uri,
 	}
 
 	if (load_uri) {
-		uri = gnome_vfs_uri_append_file_name (ctx->uri, rel_uri);	
-		
+		uri = gnome_vfs_uri_append_file_name (ctx->uri, rel_uri);
 		image = eog_image_new_uri (uri);
 		gnome_vfs_uri_unref (uri);
-		
-		if (image != NULL) {
+		if (image != NULL) {	
 			add_image_private (list, image, FALSE);
 			g_object_unref (image);
 		}
@@ -412,20 +540,30 @@ static void
 add_directory (EogImageList *list, GnomeVFSURI *uri, GnomeVFSFileInfo *info)
 {
 	DirLoadingContext ctx;
-
+	MonitorHandleContext *hctx = g_new0(MonitorHandleContext, 1);
+	
+	hctx->text_uri = gnome_vfs_uri_get_path (uri);
+	
 	g_return_if_fail (info->type == GNOME_VFS_FILE_TYPE_DIRECTORY);
 
 	ctx.uri = uri;
 	ctx.list = list;
 	ctx.info = info;
 	
+	gnome_vfs_monitor_add  (&hctx->handle, hctx->text_uri,
+		GNOME_VFS_MONITOR_DIRECTORY,
+		vfs_monitor_dir_cb,
+		list);
+
+	list->priv->monitors = g_list_append (list->priv->monitors, hctx);
+	
 	gnome_vfs_directory_visit_uri (uri,
 				       GNOME_VFS_FILE_INFO_DEFAULT |
 				       GNOME_VFS_FILE_INFO_FOLLOW_LINKS |
 				       GNOME_VFS_FILE_INFO_GET_MIME_TYPE,
 				       GNOME_VFS_DIRECTORY_VISIT_DEFAULT,
 				       directory_visit_cb,
-				       &ctx);	
+				       &ctx);
 }
 
 static void
@@ -447,7 +585,7 @@ add_regular (EogImageList *list, GnomeVFSURI *uri, GnomeVFSFileInfo *info)
 			add_image_private (list, image, FALSE);
 			g_object_unref (image);
 		}
-	}	
+	}
 }
 	
 
@@ -474,7 +612,9 @@ eog_image_list_add_uris (EogImageList *list, GList *uri_list)
 	GList *it;
 	EogImageListPrivate *priv;
 	GnomeVFSFileInfo *info;
-
+	GList *initial_image_item;
+	GnomeVFSURI *initial_uri = NULL;
+	
 	priv = list->priv;
 
 	if (uri_list == NULL) {
@@ -485,21 +625,48 @@ eog_image_list_add_uris (EogImageList *list, GList *uri_list)
 	
 	for (it = uri_list; it != NULL; it = it->next) {
 		GnomeVFSURI *uri = (GnomeVFSURI*) it->data;
-		
+
 		if (!get_uri_info (uri, info))
 			continue;
 			
 		if (info->type == GNOME_VFS_FILE_TYPE_DIRECTORY) {
 			add_directory (list, uri, info);
 		}
-		else if (info->type == GNOME_VFS_FILE_TYPE_REGULAR) {
+		else if (info->type == GNOME_VFS_FILE_TYPE_REGULAR && 
+			 g_list_length (uri_list) == 1) {
+
+			/* Store the URI for initial image assignment */
+			initial_uri = gnome_vfs_uri_dup (uri); 
+			
+			uri = gnome_vfs_uri_get_parent (uri);
+			
+			if (!get_uri_info (uri, info))
+				continue;
+			
+			add_directory (list, uri, info);
+		}	
+		else if (info->type == GNOME_VFS_FILE_TYPE_REGULAR && 
+			 g_list_length (uri_list) > 1) {
+
 			add_regular (list, uri, info);
-		}
+		}		
 	}
 
 	gnome_vfs_file_info_unref (info);
 	
 	priv->store = g_list_sort (priv->store, compare_filename_cb);
+
+	if (initial_uri) {
+		initial_image_item = g_list_find_custom (priv->store, initial_uri, 
+						    compare_uri_cb);
+		gnome_vfs_uri_unref (initial_uri);
+	}
+	else {
+		initial_image_item = g_list_first (priv->store);
+	}
+
+	if (initial_image_item != NULL)
+		priv->initial_pos = g_list_position (priv->store, initial_image_item);
 }
 
 /* ====================================================== */
@@ -557,6 +724,14 @@ eog_image_list_length (EogImageList *list)
 	return list->priv->n_images;
 }
 
+int
+eog_image_list_get_initial_pos (EogImageList *list)
+{
+	g_return_val_if_fail (EOG_IS_IMAGE_LIST (list), -1);
+	
+	return list->priv->initial_pos;
+}
+
 GnomeVFSURI*
 eog_image_list_get_base_uri (EogImageList *list)
 {
@@ -579,7 +754,7 @@ eog_image_list_print_debug (EogImageList *list)
 	
 	priv = list->priv;
 	
-	g_print ("n_images: %i/%i\n", priv->n_images, g_list_length (priv->store));
+	g_print ("n_images: %i/%u\n", priv->n_images, g_list_length (priv->store));
 	for (it = priv->store; it != NULL; it = it->next) {
 		gint width, height;
 		EogImage *img = EOG_IMAGE (it->data);
@@ -595,10 +770,17 @@ eog_image_list_print_debug (EogImageList *list)
 void
 eog_image_list_add_image (EogImageList *list, EogImage *image)
 {
+	int pos;
+	
 	g_return_if_fail (EOG_IS_IMAGE_LIST (list));
 	g_return_if_fail (EOG_IS_IMAGE (image));
 
-	add_image_private (list, image, TRUE);
+
+	pos = add_image_private (list, image, TRUE);		
+		
+	if (pos >= 0) {
+		g_signal_emit (list, eog_image_list_signals[IMAGE_ADDED], 0, image, pos);
+	}
 }
 
 void
@@ -606,15 +788,16 @@ eog_image_list_remove_image (EogImageList *list, EogImage *image)
 {
 	EogIter* iter;
 	int pos;
-	
+		
 	g_return_if_fail (EOG_IS_IMAGE_LIST (list));
 	g_return_if_fail (EOG_IS_IMAGE (image));
-
+	
 	iter = eog_image_list_get_iter_by_img (list, image);
 	if (iter == NULL) return;
 
 	pos = remove_image_private (list, iter);
 
+	
 	g_free (iter);
 
 	if (pos >= 0) {
